
(* random-access lists implemented as Braun trees *)

type seq a =
	| empty
	| node (non_empty_seq a)
and non_empty_seq a =
	left : seq a * item : a * right : seq a 

internal define para s f resume = 
	case s {
	| empty = f.empty
	| node n l r = f.node n (para n.left f) (para n.right f) }

define seq_cons x s =
	para s {
	| empty x =
		empty, x, empty 
	| node n l r x =
		n change (node (r resume n.item)) x n.left
	}
	resume x

define seq_head s =
	s.item

define seq_combine sa sb =
	para sa {
	| empty s =
		empty
	| node n l r s =
		node (n with left s with right (l resume n.right))
	}
	resume sb
	
define seq_tail s =
	seq_combine s.left s.right

define seq_lookup s n =
	para s {
	| empty i =
		error index_out_of_bounds
	| node n l r i =
		case i {
		| null = ok n.item
		| shift_zero i_div2 = r resume (i_div2 - 1)
		| shift_succ i_div2 = l resume i_div2
		}
	}
	resume n
	
define seq_non_empty f s =
	f (node s)

define seq_insert s n item =
	para s {
	| empty i =
		error index_out_of_bounds
	| node n l r i =
		case i {
		| null = ok
			(n change n.left item n.right)
		| shift_zero i_div2 = ok1
			(seq_non_empty (n with right)) (r resume (i_div2 - 1))
		| shift_succ i_div2 = ok1
			(seq_non_empty (n with left)) (l resume i_div2)
		}
	}
	resume n
	