
(* Mutable/immutable functional arrays.     
 * Element index can start at 0 but are not necessarily 
 * successive because it's not a dense array.
 * But there is a cost: access time is not constant.
 * This datatype will greatly help inductive graphs. 
 *) 

type nat_map a =
	| empty
	| node non_empty_nat_map
and non_empty_nat_map a =	
	left:nat_map a * item: option a * right:nat_map a

internal define fold2 ma mb f =
	case ma {
	| empty =
		f.empty_left mb
	| node a = 
	case mb {
	| empty =
		f.empty_right ma
	| node b =
		f.nodes a.item b.item
		(fold2 a.left b.left f)
		(fold2 a.right b.right f) }
	}
	
define nmap_insert a i d =
	cata a {
	| empty i =
		fold i { 
		| null = node(empty,some d,empty)
		| shift_zero l = node(l,none,empty)
		| shift_succ r = node(empty,none,r)
		}
	| node n l r i = n
		case i { 
		| null = node (n with item (some d))
		| shift_zero idiv2 = node (n with right (r resume idiv2))
		| shift_succ idiv2 = node (n with left  (l resume idiv2))
		}
	}
	resume i

define nmap_lookup a i =
	cata a {
	| empty i =
		error index_out_of_bounds 
	| node n l r i =
		case i { 
		| null =
			case n.item {
			| none = error not_found
			| some x = ok x }
		| shift_zero idiv2 = r resume idiv2
		| shift_succ idiv2 = l resume idiv2
		}	
	}
	resume i

define nmap_node n =
	case n.left {
	| node _ = node n
	| empty = 	
	case n.right {
	| node _ = node n
	| empty = 	
	case n.item {
	| some _ = node n
	| none = empty
	}}}	
	
define nmap_remove a x =
	cata a
	{ empty i =
		empty
	| node n l r i = 
		case i { 
		| null = nmap_node (n with item none)
		| shift_zero idiv2 = nmap_node (n with right (r resume idiv2))
		| shift_succ idiv2 = nmap_node (n with left  (l resume idiv2))
		}
	}
	resume x
	
define nmap_union f ma mb =
	fold2 ma mb {
	| empty_left r = r
	| empty_right l = l
	| nodes i j l r = node (l,option_or f i j,r)
	}
	
define nmap_intersection f ma mb =
	fold2 ma mb {
	| empty_left r = empty
	| empty_right l = empty
	| nodes i j l r = nmap_node (l,option_and f i j,r)
	}
	
define nmap_difference ma mb =
	fold2 ma mb {
	| empty_left r = empty
	| empty_right l = l
	| nodes i j l r = nmap_node (l,option_andnot i j,r)
	}
	
define nmap_subset ma mb =
	fold2 ma mb {
	| empty_left r = true
	| empty_right l = false
	| nodes i j l r = l && option_xnor i j && r)
	}
	
define nmap_equal ma mb =
	nmap_subset ma mb && nmap_subset mb ma

define nmap_update a u f =
	cata a {
	| empty u =
		empty
	| node n l r u =
		node
		( case u {
		| empty = n
		| some v =
		case v.head = 0 && v.tail = empty {
		| true =
				n with n.item
				( case n.item { none = none | some x = some (f x) } )
		| false =
				let odds  = seq_filter_map odd  div2 u empty in
				let evens = seq_filter_map even div2 u empty in
				n change
				odds
				( case n.item { none = none | some x = some (f x) } )
				evens
		}})
	} resume u
		
	