
(* Mutable/immutable set of nats.     
 * Element index can start at 0 but are not necessarily 
 * successive because it's not a dense set.
 *) 

type nat_set =
	| empty
	| node non_empty_nat_set
and non_empty_nat_set =	
	left:nat_set * item:bool * right:nat_set

internal define fold2 sa sb f =
	case sa {
	| empty =
		f.empty_left sb
	| node a = 
	case sb {
	| empty =
		f.empty_right sa
	| node b =
		f.nodes a.item b.item
		(fold2 a.left b.left f)
		(fold2 a.right b.right f) }
	}
	
define nset_insert a i =
	cata a {
	| empty i =
		fold i { 
		| null = node(empty,true,empty)
		| shift_zero l = node(l,false,empty)
		| shift_succ r = node(empty,false,r)
		}
	| node n l r i = n
		case i { 
		| null = node (n with item true)
		| shift_zero idiv2 = node (n with right (r resume idiv2))
		| shift_succ idiv2 = node (n with left  (l resume idiv2))
		}
	}
	resume i

define nset_member a i =
	cata a {
	| empty i =
		false 
	| node n l r i =
		case i { 
		| null = n.item
		| shift_zero idiv2 = r resume idiv2
		| shift_succ idiv2 = l resume idiv2
		}	
	}
	resume i

define nset_node n =
	case n.left {
	| node _ = node n
	| empty = 	
	case n.right {
	| node _ = node n
	| empty = 	
	case n.item {
	| true = node n
	| false = empty
	}}}	
	
define nset_remove a x =
	cata a
	{ empty i =
		empty
	| node n l r i = 
		case i { 
		| null = nset_node (n with item false)
		| shift_zero idiv2 = nset_node (n with right (r resume idiv2))
		| shift_succ idiv2 = nset_node (n with left  (l resume idiv2))
		}
	}
	resume x
	
define nset_union sa sb =
	fold2 sa sb {
	| empty_left r = r
	| empty_right l = l
	| nodes i j l r = node (l,i or j,r)
	}
	
define nset_intersection sa sb =
	fold2 sa sb {
	| empty_left r = empty
	| empty_right l = empty
	| nodes i j l r = nset_node (l,i && j,r)
	}
	
define nset_difference sa sb =
	fold2 sa sb {
	| empty_left r = empty
	| empty_right l = l
	| nodes i j l r = nset_node (l,i andnot j,r)
	}
	
define nset_subset sa sb =
	fold2 sa sb {
	| empty_left r = true
	| empty_right l = false
	| nodes i j l r = l && i xnor j && r)
	}
	
define nset_equal sa sb =
	nset_subset sa sb && nset_subset sb sa
		