
(* Mutable/immutable totally ordered Sets *)

type set a =
	| empty
	| node (non_empty_set a)
and non_empty_set a =
	left : set a * item : a * right : set a 

internal define map s f =
	case s {
	| empty = empty
	| node (l,x,r) = node (map l f, f.node x, map r f) }

internal define fold s f = 
	case s {
	| empty = f.empty
	| node (l,x,r) = f.node (fold l f) x (fold r f) }

internal define cata s f resume = 
	case s {
	| empty = f.empty
	| node (l,x,r) = f.node (cata l f) x (cata r f) }

internal define recu s f = 
	case s {
	| empty = f.empty
	| node n = f.node n (recu n.left f) (recu n.right f) }

internal define para s f resume = 
	case s {
	| empty = f.empty
	| node n = f.node n (para n.left f) (para n.right f) }

define para_non_empty n f resume =
	f.node n (para n.left f) (para n.right f) 						

define cardinal s =
	fold s {
	| empty = 0
	| node l _ r = l + 1 + r }

define for_all cond s =
	cata s {
	| empty = true
	| node l y r = cond y && (l resume) && (r resume)
	}
	resume
	
define set_member x s =
	cata s {
	| empty =
		false
	| node l y r =
		case (compare x y) {
		| less  =	l resume
		| more  =	r resume
		| equal =	true	
		}
	} resume
		
define set_insert x s =
	para s {
	| empty =
		empty, x, empty 
	| node n l r =
		case (compare x n.item) {
		| less  =	n with left  (node (l resume))
		| more  =	n with right (node (r resume))
		| equal =	n
		}
	} resume	

define set_min s =
	para_non_empty s {
	| empty =
		unused s
	| node n l r =
		case n.left {
		| empty = n
		| node _ = l resume
		}
	} resume

define set_remove_min s =
	para_non_empty s {
	| empty =
		unused empty
	| node n l r =
		case n.left {
		| empty = n.right
		| node _ = node (n with left (l resume))
		}
	} resume
	
define set_concat sa sb =
	case sa {
	| empty =
		sb
	| node _ =
		case sb {
		| empty =
			sa
		| node n =
			let m = set_min n in
			let r = set_remove_min n in
			node (m with left sa with right r) }
	}
		
define set_remove x s =
	para s {
	| empty =
		empty
	| node n l r =
		case (compare x n.item) {
		| less  = node (n with left  (l resume))
		| more  = node (n with right (r resume))
		| equal = set_concat n.left n.right
		}
	} resume
	
define set_split x s =
	para s {
	| empty =
		empty, false, empty
	| node n l r =
		case (compare x n.item) {
		| less =	let a,b,c = l resume in
					a, b, node (n with left c)
		| more =	let a,b,c = r resume in
					(n with right a) , b , c
		| else = n.left, true, n.right
		}
	} resume

define set_filter cond s =
	recu s {
	| empty =
		empty
	| node n l r =
		case (cond n.item) {
		| true =
			node (n with left l with right r)
		| false =
			set_concat l r 
		}
	}

define set_union sa sb =
	para sa {
	| empty s =
		s
	| node m l r s =
		node
		( case s {
		| empty =
			m
		| node n =
			let a,b,c = set_split m.item s in
			m with left (l resume a) with right (r resume c)}
		)
	} resume sb

define set_intersection sa sb =
	para sa {
	| empty s =
		empty
	| node m l r s =
		case s {
		| empty =
			empty
		| node n =
			let a,b,c = set_split m.item s in
			case b {
			| true =
				node (m with left (l resume a) with right (r resume c))
			| false =
				set_concat (l resume a) (r resume c)
			}
		}
	} resume sb

define set_difference sa sb =
	para sa {
	| empty s =
		empty
	| node m l r s =
		case s {
		| empty =
			node m
		| node n =
			let a,b,c = set_split m.item s in
			case b {
			| true =
				set_concat (l resume a) (r resume c)
			| false =
				node (m with left (l resume a) with right (r resume c))
			}
		}
	} resume sb

define set_subset sa sb =
	para sa {
	| empty s =
		true
	| node m l r s =
		case s {
		| empty =
			false
		| node n =
			case (compare m.item n.item) {
			| less =
				(l resume n.left) && set_member m.item n.left && (r resume s)
			| more =
				(l resume s) && set_member m.item n.right && (r resume n.right)
			| equal =
				(l resume n.left) && (r resume n.right)
			}
		}
	} resume sb

define set_equal sa sb =
	set_subset sa sb && set_subset sb sa
	