
(* A simple standard prelude *)

(* Booleans *)
internal type bool =
	| true
	| false

(* Total order *)
type order =
	| less
	| equal
	| more

(* Delayed evaluation *)
type pass =
	resume

define halt f x resume =
	f x
	
(* Peano naturals *)	
internal type nat =
	| zero
	| succ nat

internal define fold n f = 
	case n {
	| zero = f.zero
	| succ p = f.succ (fold p f) }

internal define recu n f = 
	case n {
	| zero = f.zero
	| succ p = f.succ p (recu p f) }

internal define para n f resume = 
	case n {
	| zero = f.zero
	| succ p = f.succ p (para p f) }
		
(* Naturals, base 2 *)	
internal type nat =
	| null
	| shift_zero nat
	| shift_succ nat

internal define fold n f = 
	case n {
	| null = f.null
	| shift_zero h = f.shift_zero (fold h f)
	| shift_succ h = f.shift_succ (fold h f) }

internal define cata n f = 
	case n {
	| null = f.null
	| shift_zero h = f.shift_zero (cata h f)
	| shift_succ h = f.shift_succ (cata h f) }

internal define recu n f = 
	case n {
	| zero = f.zero
	| shift_zero h = f.shift_zero (recu h f)
	| shift_succ h = f.shift_succ (recu h f) }

internal define para n f resume = 
	case n {
	| zero = f.zero
	| shift_zero h = f.shift_zero (para h f)
	| shift_succ h = f.shift_succ (para h f) }
 		
(* Utitity fonctions *)		
define id x = x
define id1 x y = x
define id2 x y = y
define ignore1 f x y = f x
define ignore2 f x y = f y
define compose g f x = g (f  x)
define ever x = true
define never x = false
define const k x = k
define flip f a b = f b a
define swap (a,b) = (b,a)
define left (a,b) = a
define right(a,b) = b
  	
define times n f x =
	fold n {
	| zero: x
	| succ m: f m)
	}

(* Booleans functions *)
define neg f x =
	case (f x) {
	| true = false
	| false = true }
	
define not b =
	case b {
	| true = false
	| false = true }

(* More arithmetic utilities *)

define max a b =
	case (compare a b) {
	| more = a 
	| else = b } 

define min a b =
	case (compare a b) {
	| less = a 
	| else = b } 

define even n =
	case n {
	| shift_succ _ = false
	| else = true}

define odd n =
	case n {
	| shift_succ _ = true
	| else = false}

define div2 n =
	case n {
	| null = null 
	| shift_zero h = h
	| shift_succ h = h }

(* Inclusion in interval [a,b] *)	
define between x a b =
	case (compare x a) {
	| less = false
	| else =
	case (compare x b) {
	| more = false
	| else = true }
	}
	
define power x n =
	fold n {
	| zero = 1
	| succ xm = x * xm }

define product a b = (* a >= b *)
	recu (a - b) {
	| zero = b
	| succ m fm = (succ m) * fm }

define factorial n =
	case n {
	| zero = 1
	| else = product n 1 }

(* Binomial coefficients *)		
define binomial n p =
	para n {
	| zero p = succ zero
	| succ m bm p =
		case p {
		| zero = succ zero
		| else =
			case (compare p (succ m)) {
			| equal = succ zero
			| else =
				let bmr = bm resume 
				in  bmr p + bmr (p-1) } } 
	} p

(* Catalan numbers *)	
define Catalan n =
	binomial (n+n) n / (n+1)
	
(* Finonacci numbers *)	
define Finonacci n =
	left
	( fold n {
	| zero = (0,1)
	| succ (a,b) = (b,a+b) }
	)

(* Moonscript cannot implement Ackermann *)
			