	
(* Result monad *)	
type result a b =
	| ok a
	| error b

type std_error =
	| arithmetic_overflow
	| division_by_zero
	| invalid_argument	
	| invalid_case	
	| not_found
	| index_out_of_bounds	
	| no_free_memory	
	
define ok1 f v =
  case v { 
  | ok x = ok (f x)
  | else = v }  

define result1 f v =
  case v { 
  | ok x = f x
  | else = v }  

define ok2 f v1 v2 =
  case v1 { 
  | ok x1 =
	  case v2 { 
	  | ok x2 = ok (f x1 x2)
	  | else  = v2 }  
  | else = v1 }  

define result2 f v1 v2 =
  case v1 { 
  | ok x1 =
	  case v2 { 
	  | ok x2 = f x1 x2
	  | else = v2 }  
  | else = v1 }  

(* With default value *)  	
define with_default x expr =
	case expr {
	| ok y = y
	| else = x }

(* With exception handler *)  	
define with_error_handler expr h =
	case expr {
	| ok _ = expr
	| error e = h e }

(* Safe arithmetic *)	
	
define safe_division a b =
	case b {
	| zero = error division_by_zero
	| else = ok (a / b) }
		
define safe_modulo a b =
	case b {
	| zero = error division_by_zero
	| else = ok (a mod b) }
	
define safe_division a b =
	case b {
	| zero = error division_by_zero
	| else = ok (a / b) }
		
define safe_div_mod a b =
	case b {
	| zero = error division_by_zero
	| else = ok (a/b, a mod b) }
