
(* Mutable/immutable linear sequences *)

type seq a =
	| empty
	| node (non_empty_seq a)
and non_empty_seq a =
	head : a * tail : seq a

internal define map s f =
	case s {
	| empty = empty
	| node (h,t) = node (f.node h, map t f) }
		
internal define fold s f =
	case s {
	| empty = f.empty
	| node (h,t) = f.node h (fold t f) }
		
internal define recu s f =
	case s {
	| empty = f.empty
	| node n = f.node n (recu n.tail f) }

internal bicata define sa sb f resume =
	sa sb {
	case sa {
	| empty =
		f.empty_left sb
	| node a = 
	case sb {
	| empty =
		f.empty_right sa
	| node b =
		f.nodes a b
		(bicata a.tail sb f) 
		(bicata sa b.tail f) }
	
define seq_cons x s =
	{head=x | tail=s} 

define seq_make x s =
	{head=x | tail=node s} 

define seq_singleton x =
	{head=x | tail=empty}
	
define seq_length s =
	fold s {
	| empty = 0
	| node _ n = succ n }
	}	

define seq_init s =
	case s.tail {
	| empty = empty
	| node l =
		fold l {
		| empty x = empty
		| node h t x = seq_cons x (t h) 
		} s.head
	}	

define seq_last s =
	case s.tail {
	| empty = s.head
	| node l =
		fold l {
		| empty x = x
		| node h t x = t h 
		} s.head
	}	
	
define seq_insert n x s =
	cata n {
	| zero l =
		case l {
		| empty = error index_out_of_bounds
		| node n = ok (node (n with head x)) }
	| succ p l =
		case l {
		| empty = error index_out_of_bounds
		| node n = p() n.tail}
	}
	resume s		
	
define seq_lookup n s =
	cata n {
	| zero l =
		case l {
		| empty = error index_out_of_bounds
		| node n = ok n.head }
	| succ p l =
		case l {
		| empty = error index_out_of_bounds
		| node n = p() n.tail}
	}
	resume s		
	
define seq_append l1 l2 =
	recu l1 {
	| empty = l2
	| node n t = n with tail t }
	}	 

define seq_rev_append l1 l2 =
	left (
		fold l1 {
		| empty = (l2,l1)
		| node a (b,c) =
			case c {
			| empty = (cons a b,empty)
			| node(d,e) = (d::b,e)	
		}})		
			
define seq_reverse l =
	seq_rev_append l empty

define seq_append_rev l1 l2 =
  seq_append l1 (seq_rev_append l2 empty)
	
define seq_flatten l =
	fold l {
	| empty = empty
	| node h t = seq_append h t }
	}	
	
define seq_filter_map cond f l1 l2 = 
	recu l1 {
	| empty = l2 
	| node n t =
		case cond n.head {
		| true = n change (f n.head) t
		| false = t
		}
	}	
	
define seq_filter_append cond l1 l2 = 
	recu l1 {
	| empty = l2
	| node n t =
		case cond n.head {
		| true = n with tail t
		| false = t
		}
	}	
	
define seq_filter cond l = 
	seq_filter_append cond l empty 

define seq_power l =
	fold l {
	| empty
	| node h t: seq_append t (map t | node x = node h x)
	}	
  	
define seq_range a b = (* a <= b *)
	fold (b-a) {
	| zero = {head=a | tail=empty}
	| succ n = {head=succ n.head | tail=node n}
	}
	
define seq_even_index l = 
	fold l {
	| empty n = empty
	| node h t n =
		case n {
		| shift_succ _ = t
		| else =  seq_cons h t
		} (succ n)
	} zero	
	
define seq_odd_index l = 
	fold l {
	| empty n = empty
	| node h t n =
		case n {
		| shift_succ _ = seq_cons h t
		| else = t 
		} (succ n)
	} zero	
	
(* merge two sorted lists *)
define seq_merge sa sb =
	bicata sa sb {
	| empty_left r = r
	| empty_right l = l
	| nodes a b ma mb =
		node 
		( case (compare a.head b.head) {
		| more = a with tail (ma resume)
		| else = b with tail (mb resume)
		)	
	}

		 