
(* a simple calculator *)

type arithmetic =
	| Cst of int
	| Neg of arithmetic
	| Add of binary
	| Sub of binary
	| Mul of binary
	| Div of binary
and binary =
	{left: arithmetic; right: arithmetic}
	
let rec fold a f =
	match a with
	| Cst n  -> f#cst n 
	| Neg op -> f#neg (fold op f)
	| Add op -> f#add (fold op.left f) (fold op.right f)
	| Sub op -> f#sub (fold op.left f) (fold op.right f)
	| Mul op -> f#mul (fold op.left f) (fold op.right f)
	| Div op -> f#div (fold op.left f) (fold op.right f)

let eval expr =
	fold expr (
		object
			method cst n = n
			method neg a = (~-) a
			method add a b = ( + ) a b
			method sub a b = ( - ) a b 
			method mul a b = ( * ) a b 
			method div a b = ( / ) a b
		end )
			
type ('a,'b) result =
	| Ok of 'a
	| Error of 'b 

let ok1 f v =
	match v with
	| Ok x -> Ok (f x)
	| Error _ -> v

let ok2 f v1 v2 =
	match v1,v2 with
	| Ok x1,Ok x2 -> Ok (f x1 x2) 
	| Error _, _ -> v1
	| _, Error _ -> v2

let result2 f v1 v2 =
	match v1,v2 with
	| Ok x1,Ok x2 -> f x1 x2 
	| Error _, _ -> v1
	| _, Error _ -> v2
	
let safe_division a b =  
	if b=0 then Error "Division by 0"
	else Ok (a/b)	
	
let eval expr =
	fold expr (
		object
			method cst n = Ok n
			method neg a = ok1 (~-) a
			method add a b = ok2 ( + ) a b
			method sub a b = ok2 ( - ) a b 
			method mul a b = ok2 ( * ) a b 
			method div a b = result2 safe_division a b
		end )
		
let ok1_error f v1 (cond,code) =
	match v1 with
	| Error _ -> v1
	| Ok x1 ->
		if cond x1 then Error code
		else Ok (f x1)	
		
let ok2_error f v1 v2 (cond,code) =
	match v1,v2 with
	| Error _, _ -> v1
	| _, Error _ -> v2
	| Ok x1,Ok x2 ->
		if cond x1 x2 then Error code
		else Ok (f x1 x2)	
	
let eval expr =
	fold expr (
		object
			method cst n = Ok n
			method neg a = ok1 (~-) a
			method add a b = ok2 ( + ) a b
			method sub a b = ok2 ( - ) a b 
			method mul a b = ok2 ( * ) a b 
			method div a b = ok2_error ( / ) a b ((fun a b -> b=0),"Division by 0")
		end )
		           