
(* random-access lists implemented as Braun trees *)

module type SeqConstructors =
	sig
		type 'a seq =
			'a non_empty_seq option
		and 'a non_empty_seq =
			private
			{mutable left: 'a seq; mutable item: 'a; mutable right: 'a seq}
		val empty  : 'a seq
		val make   : 'a seq -> 'a -> 'a seq -> 'a non_empty_seq
		val change : 'a non_empty_seq ->
		             'a seq -> 'a -> 'a seq -> 'a non_empty_seq
		val with_left  : 'a non_empty_seq -> 'a seq -> 'a non_empty_seq
		val with_right : 'a non_empty_seq -> 'a seq -> 'a non_empty_seq
	end


module SeqRecursors(S:SeqConstructors) =
	struct
		include S
		(* errors *)
		type ('a,'b) result =
			| Ok of 'a
			| Error of 'b
		let ok1 f v =
			match v with
			| Ok x -> Ok (f x)
			| Error _ -> v
		(* lazy paramorphism *)
		type ('a,'b) para =
			<empty: 'b;
			 node: 'a non_empty_seq -> (unit -> 'b) -> (unit -> 'b) -> 'b>
		let rec para s (case:('a,'b) para) () =
			match s with
			| None ->
					case#empty
			| Some n ->
					case#node n (para n.left case) (para n.right case)
	end

module MakeSeq(S:SeqConstructors)
	=
	struct
		(* general *)
		include
			SeqRecursors(S)
		let seq_cons x s =
			para s (
				object
					method empty x =
						make None x None						
					method node n l r x =
						change n (Some (r () n.item)) x n.left
				end ) () x
		let seq_head s =
			s.item
		let seq_combine sa sb =
			para sa (
				object
					method empty s =
						None
					method node n l r s =
						Some (with_right (with_left n s) (l()n.right))
				end ) () sb
		let seq_tail s =
			seq_combine s.left s.right
		let seq_lookup n s =
			para s (
				object
					method empty i =
						Error "index_out_of_bounds"
					method node n l r i  =
						if i = 0 then Ok n.item
						else if i land 1 = 1 then l () (i/2) 
						else r () (i/2 -1)						 
				end ) () n
		let seq_non_empty f s =
			f (Some s)
		let seq_update n item s =
			para s (
				object
					method empty i =
						Error "index_out_of_bounds"
					method node n l r i  =
						if i = 0 then Ok
							(change n n.left item n.right)
						else if i land 1 = 1 then ok1
							(seq_non_empty (with_left n)) (l () (i/2))
						else ok1
							(seq_non_empty (with_right n)) (r () (i/2 -1))
				end ) () n
	end

				
	