
(* Mutable/immutable Double-Ended-queue *)
	
module Dequeue(S:Seq)
	:
	sig
		type 'a queue =
			'a non_empty_queue option
		and 'a non_empty_queue = {
			front : 'a S.non_empty_seq;
			rear : 'a S.seq;
			}
		val empty : 'a queue
		val snoc : 'a -> 'a queue -> 'a non_empty_queue
		val cons : 'a -> 'a queue -> 'a non_empty_queue
		val head : 'a non_empty_queue -> 'a
		val tail : 'a non_empty_queue -> 'a queue
		val length: 'a queue -> int
	end
	=
	struct
		open S
		type 'a queue =
			'a non_empty_queue option
		and 'a non_empty_queue =
			{front: 'a non_empty_seq; rear: 'a seq}
		let empty =
			None
		let snoc x = function
			| None ->
				{front=cons x None; rear=None}
			| Some q ->
				{q with rear=Some(cons x q.rear)} 
		let cons x = function
			| None ->
				{front=cons x None; rear=None}
			| Some q ->	
				{q with front=make x q.front}
		let head q =
			q.front.head
		let tail q =
			match q.front.tail with
			| None ->
				( match q.rear with
				| None -> None
				| Some r -> Some{front=reverse r; rear=None} )
			| Some t -> Some{q with front=t}
		let length q =
			match q with
			| None -> 0
			| Some n -> length (Some n.front) + length n.rear
	end
