
(* a simple formal derivator *)

type function_x =
	| Var
	| Cst of int
	| Neg of function_x
	| Add of function_x * function_x
	| Mul of function_x * function_x
	| Sin of function_x 
	| Cos of function_x
	| Exp of function_x
	
let rec recu f case =
	match f with
	| Var ->
			case#var
	| Cst n ->
			case#cst n
	| Neg u ->
			case#neg u (recu u case)
	| Add (u,v) ->
			case#add (u,v) (recu u case) (recu v case)
	| Mul (u,v) ->
			case#mul (u,v) (recu u case) (recu v case)
	| Sin u ->
			case#sin u (recu u case)
	| Cos u ->
			case#cos u (recu u case)
	| Exp u ->
			case#exp u (recu u case)

let deriv f =	
	recu f (		
		object	
			method var = Cst 1
			method cst n = Cst 0
			method neg _ du = Neg du
			method add _ du dv = Add (du,dv)
			method mul (u,v) du dv = Add (Mul (du,v), Mul (u,dv))
			method sin u du = Mul (du, Cos u)
			method cos u du = Neg (Mul (du, Sin u))
			method exp u du = Mul (du, Exp u)
		end )

