
type rubik_move =
	| Initial
	| Left  of rubik_move
	| Right of rubik_move
	| Up    of rubik_move
	| Down  of rubik_move
	| Front of rubik_move
	| Back  of rubik_move

(* solution by catamorphism *)
let rec inverse acc = function
	| Initial -> acc
	| Left c  -> inverse (Right acc) c
	| Right c -> inverse (Left  acc) c
	| Up c    -> inverse (Down acc) c
	| Down c  -> inverse (Up acc) c
	| Front c -> inverse (Back acc) c
	| Back c  -> inverse (Front acc) c
let inverse = inverse Initial		

(* generation by anamorphism *)
let rec rev_append_map f l1 l2 =
	match l1 with
	| [] ->
		l2
	| h::t ->
		rev_append_map f t (
		f#left  h ::
		f#right h ::
		f#up    h ::
		f#down  h ::
		f#front h ::
		f#back  h :: l2 )
		
let rec ana n case =
	if n = 0 then [case#initial]
	else rev_append_map case (ana (n-1) case) []

let cubes n =
	ana n (
		object
			method initial = Initial
			method left  c = Left  c
			method right c = Right c
			method up    c = Up    c
			method down  c = Down  c
			method front c = Front c
			method back  c = Back  c 
		end )
	
(* solution by hylomorphism *)
exception Found of rubik_move 	

(* anamorphism part of the hylomorphism *)
let rec hylo n solved b c =
	if solved (b:rubik_move) c then
		raise (Found c);
	if n > 0 then 
		let hylo_pred = hylo (n-1) solved b in
		hylo_pred (Left c);
		hylo_pred (Right c);
		hylo_pred (Up c);
		hylo_pred (Down c);
		hylo_pred (Front c);
		hylo_pred (Back c)
		
let hylo solved b =
	let n = ref 0 in
	while true do
		hylo !n solved b Initial; incr n 
	done

(* catamorphism part of the hylomorphism *)
let solved b c =
	inverse c = b
	 
(* hylo solved (Front (Up (Right Initial)));; *)

