
[color=sienna](* a Trie is an Indexed when type 'a indexed = 'a trie *)[/color]  
[color=blue]module[/color] [color=blue]type[/color] Indexed
   =
   [color=blue]sig[/color]
      [color=blue]type[/color] 'a indexed
      [color=blue]val[/color] empty : 'a indexed
      [color=blue]val[/color] lookup: 'a indexed -> int -> 'a 
      [color=blue]val[/color] insert: 'a indexed -> int -> 'a -> 'a indexed
      [color=blue]val[/color] remove: 'a indexed -> int -> 'a indexed
      [color=blue]val[/color] choose: 'a indexed -> int
      [color=blue]val[/color] update: 'a indexed -> int list -> ('a->'a) -> 'a indexed
   [color=blue]end[/color]
   
[color=sienna](* no labels at nodes, no labels at edges *)[/color]
[color=blue]module[/color] [color=blue]type[/color] UnlabeledGraphType
   =
   [color=blue]sig[/color]          
      [color=blue]type[/color] node =
         int         
      [color=blue]type[/color] context =
         {succ: node list; pred: node list}
      [color=blue]type[/color]
         graph
      [color=blue]val[/color] empty:
         graph
      [color=blue]val[/color] cons:
         node -> context -> graph -> graph 
      [color=blue]val[/color] decons:
         graph -> node -> (context * graph)
      [color=blue]val[/color] uncons:
         graph -> (node * context * graph)
   [color=blue]end[/color]

[color=blue]module[/color] UnlabeledGraph (I:Indexed)
   =
   [color=blue]struct[/color]
      [color=blue]type[/color] node =
         int         
      [color=blue]type[/color] context =
         {succ: node list; pred: node list}
      [color=blue]type[/color] graph =
         context I.indexed
      [color=blue]let[/color] empty =
         I.empty
      [color=blue]let[/color] add_succ n c =
         {c [color=blue]with[/color] succ = n::c.succ}
      [color=blue]let[/color] add_pred n c =
         {c [color=blue]with[/color] pred = n::c.pred}
      [color=blue]let[/color] cons n c g =
         [color=blue]let[/color] g = I.insert g n c [color=blue]in[/color]
         [color=blue]let[/color] g = I.update g c.pred (add_succ n) [color=blue]in[/color]
         [color=blue]let[/color] g = I.update g c.succ (add_pred n) [color=blue]in[/color]
         g
      [color=blue]let[/color] [color=blue]rec[/color] remove_edge n = [color=blue]function[/color]
         | [] -> []
         | h::t -> [color=blue]if[/color] h = n [color=blue]then[/color] t
                   [color=blue]else[/color] h::remove_edge n t
      [color=blue]let[/color] remove_succ n c =
         {c [color=blue]with[/color] succ = remove_edge n c.succ}
      [color=blue]let[/color] remove_pred n c =
         {c [color=blue]with[/color] pred = remove_edge n c.pred}
      [color=blue]let[/color] decons g n =
         [color=blue]let[/color] c = I.lookup g n [color=blue]in[/color]
         [color=blue]let[/color] g = I.remove g n [color=blue]in[/color]
         [color=blue]let[/color] g = I.update g c.pred (remove_succ n) [color=blue]in[/color]
         [color=blue]let[/color] g = I.update g c.succ (remove_pred n) [color=blue]in[/color]
         c,g
      [color=blue]let[/color] uncons g =
         [color=blue]let[/color] n = I.choose g [color=blue]in[/color]
         [color=blue]let[/color] c,g = decons g n [color=blue]in[/color]
         n,c,g 
   [color=blue]end[/color]

      
[color=blue]module[/color] UnlabeledGraphAlgos (UGgraph:UnlabeledGraphType)
   =
   [color=blue]struct[/color]
      [color=blue]include[/color] UGgraph
      [color=sienna](* the graph functor *)
      (* yes, we can have a functor even without labels *)[/color]
      [color=blue]let[/color] [color=blue]rec[/color] map f g =
         [color=blue]if[/color] g = empty [color=blue]then[/color] empty [color=blue]else[/color] 
         [color=blue]let[/color] n,c,g = uncons g [color=blue]in[/color] cons n (f c) (map f g)
      [color=sienna](* uses map to transpose a graph *)[/color]   
      [color=blue]let[/color] transpose g =
         map ([color=blue]fun[/color] {succ=s; pred=p} -> {succ=p; pred=s}) g
      [color=sienna](* add more algos here *)[/color]
   [color=blue]end[/color]
   